from abc import ABC, ABCMeta
from fastapiwee.crud.exceptions import NotFoundExceptionHandler
import re

from fastapi.params import Depends
from starlette.responses import Response
from fastapiwee.pwpd import PwPdModelFactory
from typing import Any, Optional, Type, Union

import peewee as pw
import pydantic as pd
from fastapi import FastAPI, APIRouter


class FastAPIView(ABC):
    MODEL: Type[pw.Model]
    _RESPONSE_MODEL: Optional[pd.BaseModel] = None
    URL: str
    METHOD: str
    STATUS_CODE: int = 200

    def __init__(self):
        self._response_model = self._RESPONSE_MODEL

    def __call__(self, *args, **kwargs) -> Any:
        raise NotImplementedError

    def _get_query(self) -> pw.ModelSelect:
        return self.MODEL.select()

    @property
    def response_model(self) -> pd.BaseModel:
        if self._response_model is None:
            raise ValueError('Response model is not defined')

        return self._response_model

    def _get_api_route_params(self) -> dict:
        return {
            'path': self.URL,
            'endpoint': self,
            'methods': [self.METHOD],
            'response_model': self.response_model,
            'status_code': self.STATUS_CODE,
            'name': re.sub(r'(?<!^)(?=[A-Z])', '_', self.__class__.__name__.replace('FastAPIView', '')).lower(),
        }

    def add_to_app(self, app: Union[FastAPI, APIRouter]):
        app.add_api_route(
            **self._get_api_route_params()
        )
        if isinstance(app, FastAPI):
            NotFoundExceptionHandler.add_to_app(app)

    @classmethod
    def make_model_view(cls, model: Type[pw.Model]) -> Type['FastAPIView']:
        return type(model.__name__ + cls.__name__, (cls, ), {'MODEL': model})


class BaseReadFastAPIView(FastAPIView, metaclass=ABCMeta):
    @property
    def response_model(self):
        if self._response_model is None:
            self._response_model = PwPdModelFactory(self.MODEL).read_pd

        return self._response_model

    def _get_instance(self, pk: Any) -> Type[pw.Model]:
        return self._get_query().where(self.MODEL._meta.primary_key == pk).get()


class BaseWriteFastAPIView(BaseReadFastAPIView, metaclass=ABCMeta):
    _SERIALIZER: Optional[pd.BaseModel] = None

    def __init__(self):
        super().__init__()
        self._serializer = self._SERIALIZER
        self._obj_data = None

    @property
    def serializer(self):
        if self._serializer is None:
            self._serializer = PwPdModelFactory(self.MODEL).write_pd

        return self._serializer

    def create(self) -> pw.Model:
        return self.MODEL.create(**self._obj_data.dict())

    def update(self, pk: Any, partial: bool = False) -> pw.Model:
        instance = self._get_instance(pk)
        for name, value in self._obj_data.dict(exclude_unset=partial).items():
            setattr(instance, name, value)

        instance.save()

        return instance

    def _get_api_route_params(self) -> dict:
        def data_dependency(data: self.serializer):
            self._obj_data = data

        params = super()._get_api_route_params()
        params.update({
            'dependencies': [Depends(data_dependency)],
        })

        return params


class BaseDeleteFastAPIView(BaseReadFastAPIView):
    STATUS_CODE = 204

    @property
    def response_model(self):
        return None

    def __call__(self, pk: Any):
        self.delete(pk)

    def delete(self, pk: Any):
        instance = self._get_instance(pk)
        instance.delete_instance()

    def _get_api_route_params(self) -> dict:
        params = super()._get_api_route_params()
        del params['response_model']
        params.update({
            'response_class': Response
        })

        return params



import string
from collections.abc import Mapping
from uuid import UUID
from typing import Any, List, Optional, Type

import peewee as pw
from pydantic.main import (
    BaseConfig,
    BaseModel as PdBaseModel,
    ModelMetaclass,
    inherit_config,
)
from pydantic.utils import GetterDict


def deep_update(d: dict, u: dict) -> dict:
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def letter_hash(obj: Any):
    numbers = str(hash(obj))
    negative = numbers.startswith('-')
    numbers = numbers[negative:]

    letters = ''

    for i in range(0, len(numbers), 2):
        num = int(numbers[i:i+2])
        if num > (len(string.ascii_letters) - 1):
            letters += string.ascii_letters[int(numbers[i])]
            letters += string.ascii_letters[int(numbers[i+1])]
        else:
            letters += string.ascii_letters[num]

    return ''.join(letters)


class _FieldTranslator:
    FIELDS_MAPPING = {
        pw.IntegerField: int,
        pw.FloatField: float,
        pw.BooleanField: bool,
        pw.UUIDField: UUID,
    }

    def __init__(self, field: pw.Field, nest_fk: bool = False, all_optional: bool = False):
        self.field = field
        self.nest_fk = nest_fk
        self.all_optional = all_optional

    @property
    def pd_type(self) -> str:
        field = self.field

        if isinstance(self.field, pw.ForeignKeyField):
            if self.nest_fk:
                return PwPdModel.make_serializer(self.field.rel_model)

            field = self.field.rel_field

        for field_ancestor in field.__class__.__mro__:
            pd_type = self.FIELDS_MAPPING.get(field_ancestor)
            if pd_type is not None:
                break
        else:
            pd_type = str

        if not self.is_required or self.all_optional:
            pd_type = Optional[pd_type]

        return pd_type

    @property
    def is_required(self) -> bool:
        return self.field.primary_key or (not self.field.null)


class PwPdMeta(ModelMetaclass):
    def __new__(mcs, cls_name, bases, namespace, **kwargs):
        config = namespace.get('Config', BaseConfig)
        for base in reversed(bases):
            if issubclass(base, PdBaseModel) and base != PdBaseModel:
                config = inherit_config(base.__config__, config)

        # retrieve config values
        pw_model = getattr(config, 'pw_model', None)

        if pw_model is None:
            return super().__new__(mcs, cls_name, bases, namespace, **kwargs)
            # return cls

        pw_fields = set(getattr(
            config,
            'pw_fields',
            set(pw_model._meta.fields.keys()) | set(f.backref for f in pw_model._meta.backrefs.keys())
        ))
        pw_exclude = set(getattr(config, 'pw_exclude', set()))
        exclude_pk = getattr(config, 'pw_exclude_pk', False)
        nest_fk = getattr(config, 'pw_nest_fk', False)
        nest_backrefs = getattr(config, 'pw_nest_backrefs', False)
        all_optional = getattr(config, 'pw_all_optional', False)

        allowed_fields = pw_fields - pw_exclude

        # collect peewee model fields
        fields = dict()
        for name, field in pw_model._meta.fields.items():
            if any((
                name not in allowed_fields,
                name in namespace or name in namespace.get('__annotations__', {}),
                exclude_pk and field.primary_key,
            )):
                continue

            if isinstance(field, pw.ForeignKeyField) and not nest_fk:
                name += '_id'

            fields[name] = _FieldTranslator(field, nest_fk, all_optional)

        # collect backrefs
        if nest_backrefs:
            for field, model in pw_model._meta.backrefs.items():
                if field.backref not in allowed_fields:
                    continue

                fields[field.backref] = List[PwPdModel.make_serializer(model)]

        namespace_new = {'__annotations__': {}}  # create new namespace to keep peewee fields first in order
        for name, type_ in fields.items():
            if isinstance(type_, _FieldTranslator):
                value = type_.field.default
                annotation = type_.pd_type
            else:
                value = None
                annotation = type_

            if value is not None:
                namespace_new[name] = value

            namespace_new['__annotations__'][name] = annotation

        namespace_new = deep_update(namespace_new, namespace)

        return super().__new__(mcs, cls_name, bases, namespace_new, **kwargs)


class PwPdGetterDict(GetterDict):
    def get(self, key: Any, default: Any) -> Any:
        res = getattr(self._obj, key, default)
        if isinstance(res, pw.ModelSelect):  # handle backrefs
            return list(res)
        return res


class PwPdModel(PdBaseModel, metaclass=PwPdMeta):
    __CACHE = dict()

    class Config:
        extra = 'forbid'
        orm_mode = True
        getter_dict = PwPdGetterDict

    @classmethod
    def make_serializer(cls, model: Type[pw.Model], **config_values) -> Type['PwPdModel']:
        name = model.__name__ + cls.__name__ + (letter_hash(repr(config_values)) if config_values else '')

        if name not in cls.__CACHE:
            class Config:
                pw_model = model

            for key, value in config_values.items():
                if key == 'pw_model':
                    raise ValueError('`pw_model` can not be overriden with config values, use `model` argument')

                setattr(Config, key, value)

            cls.__CACHE[name] = type(name, (cls, ), {'Config': Config})

        return cls.__CACHE[name]


class PwPdWriteModel(PwPdModel):
    """Shortcut for write model config"""
    class Config:
        pw_exclude_pk = True
        pw_nest_fk = False
        pw_nest_backrefs = False


class PwPdPartUpdateModel(PwPdWriteModel):
    """Shortcut for update model config (nothing is required)"""
    class Config:
        pw_all_optional = True


class PwPdModelFactory:
    def __init__(self, model: Type[pw.Model]):
        self._model = model
        self._read_pd = None
        self._write_pd = None

    @property
    def model(self):
        return self._model

    @property
    def read_pd(self):
        if self._read_pd is None:
            self._read_pd = PwPdModel.make_serializer(
                self._model
            )

        return self._read_pd

    @property
    def write_pd(self):
        if self._write_pd is None:
            self._write_pd = PwPdWriteModel.make_serializer(
                self._model,
            )

        return self._write_pd



from fastapiwee.crud.exceptions import NotFoundExceptionHandler
import logging
import re
from typing import Type, List, Optional

import peewee as pw
from fastapi import APIRouter, FastAPI

from fastapiwee.crud.base import FastAPIView
from fastapiwee.crud.views import (
    CreateFastAPIView,
    DeleteFastAPIView,
    ListFastAPIView,
    PartialUpdateFastAPIView,
    RetrieveFastAPIView,
    UpdateFastAPIView
)


class BaseFastAPIViewSet:
    VIEWS: Optional[List[FastAPIView]] = None

    def __init__(self, views: Optional[List[FastAPIView]] = None):
        if self.VIEWS is not None and views is not None:
            logging.warning('`VIEWS` class-level constant variable will be ignored, '
                            'since `views` argument is set on initialization.')
        self._views = views or self.VIEWS
        assert self._views, 'Views must be not null nor empty. ' \
                            'Either define `VIEWS` class-level constant variable or `views` argument on initialization.'

        self._views = [view() for view in views]  # initialize views

        self._router = None

    @property
    def router(self):
        if self._router is None:
            self._router = APIRouter(**self._get_api_router_params())
            for view in self._views:
                view.add_to_app(self._router)

        return self._router

    def _get_api_router_params(self):
        return dict()

    def add_to_app(self, app: FastAPI):
        app.include_router(self.router)
        NotFoundExceptionHandler.add_to_app(app)


class AutoFastAPIViewSet(BaseFastAPIViewSet):
    _ACTIONS_MAP = {
        'retrieve': RetrieveFastAPIView,
        'list': ListFastAPIView,
        'create': CreateFastAPIView,
        'update': UpdateFastAPIView,
        'part_update': PartialUpdateFastAPIView,
        'delete': DeleteFastAPIView,
    }

    def __init__(
        self,
        model: Type[pw.Model],
        app: FastAPI,
        actions: set = ('retrieve', 'list', 'create', 'update', 'part_update', 'delete'),
    ):
        actions = set(actions)
        self.model = model
        super().__init__(list(self._make_views(actions)))
        self.add_to_app(app)

    def _get_api_router_params(self):
        params = super()._get_api_router_params()
        params.update({
            'prefix': '/' + re.sub(r'(?<!^)(?=[A-Z])', '_', self.model.__name__).lower()  # model name to snake_case
        })

        return params

    def _make_views(self, actions: set):
        for action_name in actions:
            action_view = self._ACTIONS_MAP[action_name]
            model_view = action_view.make_model_view(self.model)

            yield model_view



from fastapiwee.pwpd import PwPdPartUpdateModel
from typing import Any, List

from fastapiwee.crud.base import (BaseDeleteFastAPIView, BaseReadFastAPIView, BaseWriteFastAPIView)


# Read
class RetrieveFastAPIView(BaseReadFastAPIView):
    METHOD = 'GET'
    URL = '/{pk}/'

    def __call__(self, pk: Any):
        return self._get_instance(pk)


# List
class ListFastAPIView(BaseReadFastAPIView):
    METHOD = 'GET'
    URL = '/'

    def __call__(self):
        return list(self._get_query())

    @property
    def response_model(self):
        return List[super().response_model]


# Create
class CreateFastAPIView(BaseWriteFastAPIView):
    METHOD = 'POST'
    URL = '/'
    STATUS_CODE = 201

    def __call__(self):
        return self.create()


# Update
class UpdateFastAPIView(BaseWriteFastAPIView):
    METHOD = 'PUT'
    URL = '/{pk}/'

    def __call__(self, pk: Any):
        return self.update(pk)


# Partial update
class PartialUpdateFastAPIView(BaseWriteFastAPIView):
    METHOD = 'PATCH'
    URL = '/{pk}/'

    def __call__(self, pk: Any):
        return self.update(pk, partial=True)

    @property
    def serializer(self):
        if self._serializer is None:
            self._serializer = PwPdPartUpdateModel.make_serializer(self.MODEL)

        return self._serializer


# Delete
class DeleteFastAPIView(BaseDeleteFastAPIView):
    METHOD = 'DELETE'
    URL = '/{pk}/'



from abc import ABC
from typing import Type

import peewee as pw
from fastapi import FastAPI, Request, Response
from fastapi.responses import JSONResponse


class ExceptionHandler(ABC):
    EXCEPTION: Type[Exception]

    def __call__(self, request: Request, exc: Exception) -> Response:
        raise NotImplementedError

    @classmethod
    def add_to_app(cls, app: FastAPI):
        app.add_exception_handler(cls.EXCEPTION, cls())


class NotFoundExceptionHandler(ExceptionHandler):
    EXCEPTION = pw.DoesNotExist

    def __call__(self, request: Request, exc: Exception) -> Response:
        return JSONResponse(
            status_code=404,
            content={
                'msg': 'Instance not found',
                'type': 'not_found',
            },
        )



import random
import string
from unittest import TestCase

import peewee as pw
import pydantic as pd
from playhouse.shortcuts import model_to_dict

from fastapiwee.pwpd import PwPdMeta, PwPdModel, PwPdModelFactory

DB = pw.SqliteDatabase(':memory:')


class ParentTestModel(pw.Model):
    id = pw.AutoField()
    text = pw.TextField()

    class Meta:
        database = DB


class TestModel(pw.Model):
    id = pw.AutoField()
    text = pw.TextField()
    number = pw.IntegerField(null=True)
    is_test = pw.BooleanField(default=True)
    related = pw.ForeignKeyField(ParentTestModel, backref='test_models')

    class Meta:
        database = DB


class ChildTestModel(pw.Model):
    id = pw.AutoField()
    test = pw.ForeignKeyField(TestModel, backref='childs')

    class Meta:
        database = DB


def create_dummy_data(amount=10, childs=5):
    DB.create_tables((ParentTestModel, TestModel, ChildTestModel))

    rel_tm = ParentTestModel.create(
        text='Parent Test',
    )

    for _ in range(amount):
        tm = TestModel.create(
            text=f'Test {"".join(random.choices(string.ascii_letters, k=8))}',
            number=random.randint(0, 1e10),
            is_test=bool(random.getrandbits(1)),
            related=rel_tm,
        )

        for _ in range(childs):
            ChildTestModel.create(
                test=tm,
            )


class PwPdMetaTestCase(TestCase):
    def test_creation(self):
        class TestModelSerializer(pd.BaseModel, metaclass=PwPdMeta):
            class Config:
                pw_model = TestModel
                pw_fields = {'id', 'text', 'number', 'related', 'childs'}
                pw_exclude = {'text'}
                pw_nest_fk = True
                pw_nest_backrefs = True

        for field_in_check in {'id', 'number', 'related', 'childs'}:
            self.assertIn(field_in_check, TestModelSerializer.__fields__)

        for field_not_in_check in {'is_test', 'text'}:
            self.assertNotIn(field_not_in_check, TestModelSerializer.__fields__)

        # FK field nested
        self.assertTrue(issubclass(TestModelSerializer.__fields__['related'].type_, PwPdModel))

        # Backref type is List[Serializer]
        self.assertIs(TestModelSerializer.__fields__['childs'].shape, pd.fields.SHAPE_LIST)
        self.assertTrue(issubclass(TestModelSerializer.__fields__['childs'].type_, PwPdModel))

        class TestModelSerializer(pd.BaseModel, metaclass=PwPdMeta):
            class Config:
                pw_model = TestModel
                pw_exclude_pk = True
                pw_nest_fk = False
                pw_nest_backrefs = False

        # All fields are in serializer by default
        model_fields = set(TestModel._meta.fields.keys())
        model_fields.remove('id')  # pw_exclude_pk = True
        model_fields.remove('related')  # pw_nest_fk = False
        model_fields.add('related_id')

        for field_in_check in model_fields:
            self.assertIn(field_in_check, TestModelSerializer.__fields__)

        # FK field not nested, just an ID
        self.assertIs(TestModelSerializer.__fields__['related_id'].type_, int)

        # Backref not included
        self.assertNotIn('childs', TestModelSerializer.__fields__)


class PwPdModelTestCase(TestCase):
    def setUp(self):
        self.tm_amount = 3
        self.child_tm_amount = 4
        create_dummy_data()

    def tearDown(self):
        DB.drop_tables((ParentTestModel, TestModel, ChildTestModel))

    def test_serialization(self):
        class TestModelSerializer(PwPdModel):
            class Config:
                pw_model = TestModel
                pw_nest_fk = True
                pw_nest_backrefs = True

        test_model = TestModel.select().order_by('?').first()
        test_model_dict = model_to_dict(test_model)
        serialized = TestModelSerializer.from_orm(test_model)
        serialized_dict = serialized.dict()

        # check values equal
        for field_check in TestModel._meta.fields.keys():
            self.assertEqual(test_model_dict[field_check], serialized_dict[field_check])

        # check backref
        self.assertEqual(test_model.childs.count(), len(serialized_dict['childs']))

    def test_make_serializer(self):
        serializer = PwPdModel.make_serializer(TestModel)
        test_model = TestModel.select().order_by('?').first()
        test_model_dict = model_to_dict(test_model)
        test_model_dict['related_id'] = test_model_dict.pop('related')['id']  # FK no nested, just ID
        serialized = serializer.from_orm(test_model)
        serialized_dict = serialized.dict()

        # check values equal
        for field_check in test_model_dict.keys():
            self.assertEqual(test_model_dict[field_check], serialized_dict[field_check])

        serializer = PwPdModel.make_serializer(TestModel, pw_fields={'id'})
        serialized = serializer.from_orm(test_model)

        self.assertListEqual(['id'], list(serialized.dict().keys()))

        with self.assertRaises(ValueError):
            PwPdModel.make_serializer(TestModel, pw_model=ParentTestModel)


class PwPdModelFactoryTestCase(TestCase):
    def setUp(self):
        create_dummy_data()
        self.test_model_pwpd = PwPdModelFactory(TestModel)

    def test_read_pd(self):
        test_model = TestModel.select().order_by('?').first()
        test_model_dict = model_to_dict(test_model)
        test_model_dict['related_id'] = test_model_dict.pop('related')['id']  # FK no nested, just ID
        serialized = self.test_model_pwpd.read_pd.from_orm(test_model)
        serialized_dict = serialized.dict()

        # check values equal
        for field_check in test_model_dict.keys():
            self.assertEqual(test_model_dict[field_check], serialized_dict[field_check])

    def test_write_pd(self):
        valid_data = {
            'text': 'Cucumber',
            'number': 300,
            'is_test': False,
            'related_id': ParentTestModel.select().first().id,
        }

        self.test_model_pwpd.write_pd(**valid_data)

        minimal_valid_data = {
            'text': 'Cucumber',
            'related_id': ParentTestModel.select().first().id,
        }
        expected_data = {
            'text': 'Cucumber',
            'number': None,
            'is_test': True,
            'related_id': ParentTestModel.select().first().id,
        }

        data = self.test_model_pwpd.write_pd(**minimal_valid_data)

        self.assertDictEqual(data.dict(), expected_data)

        invalid_data = {
            'id': 1,
            'text': 23,
            'is_test': None,
        }
        expected_errors = {
            'id': 'value_error.extra',
            'is_test': 'type_error.none.not_allowed',
            'related_id': 'value_error.missing',
        }

        with self.assertRaises(pd.ValidationError) as raised:
            self.test_model_pwpd.write_pd(**invalid_data)

        raised_errors = dict()
        for error in raised.exception.errors():
            raised_errors[error['loc'][0]] = error['type']

        self.assertDictEqual(raised_errors, expected_errors)



import setuptools


with open("README.md", "r") as readme:
    long_description = readme.read()

with open("VERSION", "r") as version_f:
    version = version_f.read()

setuptools.setup(
    name="FastAPIwee",
    version=version,
    author="German Gensetskyi",
    author_email="Ignis2497@gmail.com",
    description="FastAPIwee - FastAPI + PeeWee = <3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Ignisor/FastAPIwee",
    project_urls={
        'Documentation': 'https://fastapiwee.qqmber.wtf',
    },
    packages=setuptools.find_packages(exclude=('tests', )),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'fastapi==0.63.0',
        'pydantic==1.8.1',
        'peewee==3.14.4',
    ],
)
